
from axelrod import Player


class AntiCycler(Player):
    """
    A player that follows a sequence of plays that contains no cycles:
    C CD CCD CCCD CCCCD CCCCCD ...
    """

    name = 'AntiCycler'
    memory_depth = float('inf')

    def __init__(self):
        Player.__init__(self)
        self.cycle_length = 1
        self.cycle_counter = 0

    def strategy(self, opponent):
        if self.cycle_counter < self.cycle_length:
            self.cycle_counter += 1
            return 'C'
        else:
            self.cycle_length += 1
            self.cycle_counter = 0
            return 'D'

    def reset(self):
        Player.reset(self)
        self.cycle_length = 1
        self.cycle_counter = 0


class Cycler(Player):
    """A player that repeats a given sequence indefinitely."""

    name = 'Cycler'
    memory_depth = 1

    def __init__(self, cycle="CCD"):
        Player.__init__(self)
        self.cycle = cycle
        self.name += " " + cycle
        self.memory_depth = len(cycle)

    def strategy(self, opponent):
        curent_round = len(self.history)
        index = curent_round % len(self.cycle)
        return self.cycle[index]


class CyclerCCD(Cycler):
    def __init__(self, cycle="CCD"):
        Cycler.__init__(self, cycle=cycle)


class CyclerCCCD(Cycler):
    def __init__(self, cycle="CCCD"):
        Cycler.__init__(self, cycle=cycle)


class CyclerCCCCCD(Cycler):
    def __init__(self, cycle="CCCCCD"):
        Cycler.__init__(self, cycle=cycle)



import math
import csv
from .eigen import *

try:
    # Python 2
    from StringIO import StringIO
except ImportError:
    # Python 3
    from io import StringIO


def median(lst):
    lst = sorted(lst)
    if len(lst) < 1:
            return None
    if len(lst) % 2 == 1:
            return lst[((len(lst)+1) // 2)-1]
    if len(lst) % 2 == 0:
            return float(sum(lst[(len(lst) // 2)-1:(len(lst) // 2)+1]))/2.0


class ResultSet(object):
    """A class to hold the results of a tournament."""

    def __init__(self, players, turns, repetitions, outcome,
                 with_morality=True):
        """
        Args:
            players (list): a list of player objects.
            turns (int): the number of turns per interaction.
            repetitions (int): the number of time the round robin was repeated.
            outcome (dict): returned from the RoundRobin class and containing
                various sets of results for processing by this class.
            with_morality (bool): a flag to determine whether morality metrics
                should be calculated.
        """
        self.players = players
        self.nplayers = len(players)
        self.turns = turns
        self.repetitions = repetitions
        self.outcome = outcome
        self.results = self._results(outcome)
        self.scores = None
        self.normalised_scores = None
        self.ranking = None
        self.ranked_names = None
        self.payoff_matrix = None
        self.cooperation = None
        self.normalised_cooperation = None
        self.vengeful_cooperation = None
        self.cooperating_rating = None
        self.good_partner_matrix = None
        self.good_partner_rating = None
        self.eigenjesus_rating = None
        self.eigenmoses_rating = None
        if 'payoff' in self.results:
            self.scores = self._scores(self.results['payoff'])
            self.normalised_scores = self._normalised_scores(self.scores)
            self.ranking = self._ranking(self.scores)
            self.ranked_names = self._ranked_names(self.ranking)
            self.payoff_matrix, self.payoff_stddevs = (
                self._payoff_matrix(self.results['payoff']))
        if 'cooperation' in self.results and with_morality:
            self.cooperation = self._cooperation(self.results['cooperation'])
            self.normalised_cooperation = (
                self._normalised_cooperation(self.cooperation))
            self.vengeful_cooperation = (
                self._vengeful_cooperation(self.normalised_cooperation))
            self.cooperating_rating = self._cooperating_rating(self.cooperation)
            self.good_partner_matrix = (
                self._good_partner_matrix(self.results['cooperation']))
            self.good_partner_rating = (
                self._good_partner_rating(self.good_partner_matrix))
            self.eigenjesus_rating = (
                self._eigenvector(self.normalised_cooperation))
            self.eigenmoses_rating = (
                self._eigenvector(self.vengeful_cooperation))

    @property
    def _null_results_matrix(self):
        """
        Returns:
            A null matrix (i.e. fully populated with zero values) using
            lists of the form required for the results dictionary.

            i.e. one row per player, containing one element per opponent (in order
            of player index) which lists values for each repetition.
        """
        plist = list(range(self.nplayers))
        replist = list(range(self.repetitions))
        return [[[0 for r in replist] for j in plist] for i in plist]

    @property
    def _null_matrix(self):
        """
        Returns:
            A null n by n matrix where n is the number of players.
        """
        plist = list(range(self.nplayers))
        return [[0 for j in plist] for i in plist]

    def _results(self, outcome):
        """
        Args:
            outcome(dict): the outcome dictionary, in which the values are
                lists of the form:

                    [
                        [[a, b, c], [d, e, f], [g, h, i]],
                        [[j, k, l], [m, n, o], [p, q, r]],
                    ]

                i.e. one row per repetition, containing one element per player,
                which lists values for each opponent in order of player index.

        Returns:
            A results dictionary, in which the values are lists of
            the form:

                [
                    [[a, j], [b, k], [c, l]],
                    [[d, m], [e, n], [f, o]],
                    [[g, p], [h, q], [i, r]],
                ]

            i.e. one row per player, containing one element per opponent (in order
            of player index) which lists values for each repetition.
        """
        results = {}
        for result_type, result_list in outcome.items():
            matrix = self._null_results_matrix
            for index, result_matrix in enumerate(result_list):
                for i in range(len(self.players)):
                    for j in range(len(self.players)):
                        matrix[i][j][index] = result_matrix[i][j]
                results[result_type] = matrix
        return results

    def _scores(self, payoff):
        """
        Args:
            payoff (list): a matrix of the form:

                [
                    [[a, j], [b, k], [c, l]],
                    [[d, m], [e, n], [f, o]],
                    [[g, p], [h, q], [i, r]],
                ]

            i.e. one row per player, containing one element per opponent (in
            order of player index) which lists payoffs for each repetition.

        Returns:
            A scores matrix of the form:

                [
                    [a + b + c, j + k + l],
                    [d + e + f, m + n+ o],
                    [h + h + i, p + q + r],
                ]

            i.e. one row per player which lists the total score for each
            repetition.

        In Axelrod's original tournament, there were no self-interactions
        (e.g. player 1 versus player 1) and so these are also excluded from the
        scores here by the condition on ip and ires.
        """
        scores = []
        for ires, res in enumerate(payoff):
            scores.append([])
            for irep in range(self.repetitions):
                scores[-1].append(0)
                for ip in range(self.nplayers):
                    if ip != ires:
                        scores[-1][-1] += res[ip][irep]
        return scores

    def _normalised_scores(self, scores):
        """
        Args:
            scores (list): the scores matrix (S)

        Returns:
            A normalised scores matrix (N) such that:

                N = S / t

            where t is the total number of turns played per repetition for a given
            player excluding self-interactions.
        """
        normalisation = self.turns * (self.nplayers - 1)
        return [
            [1.0 * s / normalisation for s in r] for r in scores]

    def _ranking(self, scores):
        """
        Args:
            scores (list): the scores matrix (S)

        Returns:
            A list of players (their index within the players list rather than
            a player instance) ordered by median score
        """
        ranking = sorted(
            range(self.nplayers),
            key=lambda i: -median(scores[i]))
        return ranking

    def _ranked_names(self, ranking):
        """
        Args:
            ranking (list): A list of player index numbers

        Returns:
             A list of players names sorted by their ranked order.
        """
        ranked_names = [str(self.players[i]) for i in ranking]
        return ranked_names

    def _payoff_matrix(self, payoff):
        """
        Args:
            payoff (list): a matrix of the form:

                [
                    [[a, j], [b, k], [c, l]],
                    [[d, m], [e, n], [f, o]],
                    [[g, p], [h, q], [i, r]],
                ]

            i.e. one row per player, containing one element per opponent (in
            order of player index) which lists payoffs for each repetition.

        Returns:
            A per-turn averaged payoff matrix and its stddevs.
        """
        averages = []
        stddevs = []
        for res in payoff:
            averages.append([])
            stddevs.append([])
            for s in res:
                perturn = [1.0 * rep / self.turns for rep in s]
                avg = sum(perturn) / self.repetitions
                dev = math.sqrt(
                    sum([(avg - pt)**2 for pt in perturn]) / self.repetitions)
                averages[-1].append(avg)
                stddevs[-1].append(dev)
        return averages, stddevs

    def _cooperation(self, results):
        """
        Args:
            results (list): a matrix of the form:

                [
                    [[a, j], [b, k], [c, l]],
                    [[d, m], [e, n], [f, o]],
                    [[g, p], [h, q], [i, r]],
                ]

            i.e. one row per player, containing one element per opponent (in
            order of player index) which lists cooperation values for each
            repetition.

        Returns:
            The cooperation matrix (C) of the form:

                [
                    [[a + j], [b + k], [c + l]],
                    [[d + m], [e + n], [f + o]],
                    [[g + p], [h + q], [i + r]],
                ]

            i.e. an n by n matrix where n is the number of players. Each row (i)
            andcolumn (j) represents an individual player and the the value Cij
            is the number of times player i cooperated against opponent j.
        """
        return[[sum(element) for element in row] for row in results]

    def _normalised_cooperation(self, cooperation):
        """
        Args:
            cooperation (list): the cooperation matrix (C)

        Returns:
            A matrix (N) such that:

                N = C / t

            where t is the total number of turns played in the tournament.
        """
        turns = self.turns * self.repetitions
        return[
            [1.0 * element / turns for element in row]
            for row in cooperation]

    def _vengeful_cooperation(self, cooperation):
        """
        Args:
            cooperation(list): A cooperation matrix (C)

        Returns:
            A matrix (D) such that:

                Dij = 2(Cij -0.5)
        """
        return [[2 * (element - 0.5) for element in row] for row in cooperation]

    def _cooperating_rating(self, cooperation):
        """
        Args:
            cooperation (list): the cooperation matrix

        Returns:
            a list of cooperation rates ordered by player index"""
        total_turns = self.turns * self.repetitions * self.nplayers
        return [1.0 * sum(row) / total_turns for row in cooperation]

    def _good_partner_matrix(self, results):
        """
        Args:
            results (list): cooperation results matrix of the form:

                [
                    [[a, j], [b, k], [c, l]],
                    [[d, m], [e, n], [f, o]],
                    [[g, p], [h, q], [i, r]],
                ]

            i.e. one row per player, containing one element per opponent (in
            order of player index) which lists cooperation values for each
            repetition.

        Returns:
            The good partner matrix (P) of the form:

                [
                    [0, 0 + (1 if b >= d) + (1 if k >= m), 0 + (1 if c >= g) + (1 if l >= p) ],
                    [0 + (1 if e >= g) + (1 if n >= p), 0, 0 + (1 if f >= h) + (1 if o >= q)],
                    [0 + (1 if g >= c) + (1 if p >= l), 0 + (1 if h >= f) + (1 if q >= o), 0]
                ]

            i.e. an n by n matrix where n is the number of players. Each row (i)
            and column (j) represents an individual player and the the value Pij
            is the sum of the number of repetitions where player i cooperated as
            often or more than opponent j.
        """
        matrix = self._null_matrix
        for r in range(self.repetitions):
            for i in range(self.nplayers):
                for j in range(self.nplayers):
                    if i != j and results[i][j][r] >= results[j][i][r]:
                        matrix[i][j] += 1
        return matrix

    @property
    def _interactions(self):
        """
        Returns:
            the number of interactions between players excluding
            self-interactions.
        """
        return self.repetitions * (self.nplayers - 1)

    def _good_partner_rating(self, good_partner):
        """
        Args:
            good_partner (list): the good partner matrix

        Returns:
            a list of good partner ratings ordered by player index.
        """
        return [1.0 * sum(row) / self._interactions for row in good_partner]

    def _eigenvector(self, cooperation):
        """
        Args:
            cooperation (list): a cooperation matrix

        Returns:
            the principal eigenvector of the cooperation matrix as a list.
        """
        eigenvector, eigenvalue = principal_eigenvector(cooperation, 1000, 1e-3)
        return eigenvector.tolist()

    def csv(self):
        csv_string = StringIO()
        header = ",".join(self.ranked_names) + "\n"
        csv_string.write(header)
        writer = csv.writer(csv_string, lineterminator="\n")
        for irep in range(self.repetitions):
            data = [self.normalised_scores[rank][irep]
                    for rank in self.ranking]
            writer.writerow(list(map(str, data)))
        return csv_string.getvalue()



from axelrod import Player


class Prober(Player):
    """
    Plays D, C, C initially. Defects forever if opponent cooperated in moves 2
    and 3. Otherwise plays TFT.
    """

    name = 'Prober'
    memory_depth = float('inf')  # Long memory

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 0:
            return 'D'
        if turn == 1:
            return 'C'
        if turn == 2:
            return 'C'
        if turn > 2:
            if opponent.history[1: 3] == ['C', 'C']:
                return 'D'
            else:
                # TFT
                return 'D' if opponent.history[-1:] == ['D'] else 'C'


class Prober2(Player):
    """
    Plays D, C, C initially. Cooperates forever if opponent played D then C
    in moves 2 and 3. Otherwise plays TFT.
    """

    name = 'Prober 2'
    memory_depth = float('inf')  # Long memory

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 0:
            return 'D'
        if turn == 1:
            return 'C'
        if turn == 2:
            return 'C'
        if turn > 2:
            if opponent.history[1: 3] == ['D', 'C']:
                return 'C'
            else:
                # TFT
                return 'D' if opponent.history[-1:] == ['D'] else 'C'


class Prober3(Player):
    """
    Plays D, C initially. Defects forever if opponent played C in moves 2.
    Otherwise plays TFT.
    """

    name = 'Prober 3'
    memory_depth = float('inf')  # Long memory

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 0:
            return 'D'
        if turn == 1:
            return 'C'
        if turn > 1:
            if opponent.history[1] == 'C':
                return 'D'
            else:
                # TFT
                return 'D' if opponent.history[-1:] == ['D'] else 'C'


class HardProber(Player):
    """
    Plays D, D, C, C initially. Defects forever if opponent cooperated in moves
    2 and 3. Otherwise plays TFT.
    """

    name = 'Hard Prober'
    memory_depth = float('inf')  # Long memory

    def strategy(self, opponent):
        turn = len(self.history)
        if turn == 0:
            return 'D'
        if turn == 1:
            return 'D'
        if turn == 2:
            return 'C'
        if turn == 3:
            return 'C'
        if turn > 3:
            if opponent.history[1: 3] == ['C', 'C']:
                return 'D'
            else:
                # TFT
                return 'D' if opponent.history[-1:] == ['D'] else 'C'



import random

from axelrod import Player, Game

"""IPD Strategies: http://www.prisoners-dilemma.com/strategies.html"""


class WinStayLoseShift(Player):
    """Win-Stay Lose-Shift, also called Pavlov."""

    name = 'Win-Stay Lose-Shift'
    memory_depth = 1  # Four-Vector = (1,0,0,1)

    def __init__(self, initial='C'):
        Player.__init__(self)
        self.response_dict = {
            ('C', 'C'): 'C',
            ('C', 'D'): 'D',
            ('D', 'C'): 'D',
            ('D', 'D'): 'C',
        }
        self._initial = initial
        self.stochastic = False

    def strategy(self, opponent):
        """Switches if it doesn't get the best payout, traditionally equivalent
        to a Memory one strategy of [1,0,0,1], but this implementation does not
        require random draws."""
        if not opponent.history:
            return self._initial
        last_round = (self.history[-1], opponent.history[-1])
        return self.response_dict[last_round]


class MemoryOnePlayer(Player):
    """Uses a four-vector for strategies based on the last round of play,
    (P(C|CC), P(C|CD), P(C|DC), P(C|DD)), defaults to Win-Stay Lose-Shift.
    Intended to be used as an abstract base class or to at least be supplied
    with a initializing four_vector."""

    name = 'Generic Memory One Player'
    memory_depth = 1

    def __init__(self, four_vector, initial='C'):
        Player.__init__(self)
        self._four_vector = dict(zip([('C', 'C'), ('C', 'D'), ('D', 'C'), ('D', 'D')], map(float, four_vector)))
        self._initial = initial
        self.stochastic = False
        for x in set(four_vector):
            if x != 0 and x != 1:
                self.stochastic = True

    def strategy(self, opponent):
        if not len(opponent.history):
            return self._initial
        # Determine which probability to use
        p = self._four_vector[(self.history[-1], opponent.history[-1])]
        # Draw a random number in [0,1] to decide
        r = random.random()
        if r < p:
            return 'C'
        return 'D'


class GTFT(MemoryOnePlayer):
    """Generous Tit-For-Tat Strategy."""

    name = 'Generous Tit-For-Tat'

    def __init__(self, p=None):
        (R, P, S, T) = Game().RPST()
        if not p:
            p = min(1 - float(T - R) / (R - S), float(R - P) / (T - P))
        four_vector = [1, p, 1, p]
        super(self.__class__, self).__init__(four_vector)


class StochasticCooperator(MemoryOnePlayer):
    """Stochastic Cooperator, http://www.nature.com/ncomms/2013/130801/ncomms3193/full/ncomms3193.html."""

    name = 'Stochastic Cooperator'

    def __init__(self):
        four_vector = (0.935, 0.229, 0.266, 0.42)
        super(self.__class__, self).__init__(four_vector)


class StochasticWSLS(MemoryOnePlayer):
    """Stochastic WSLS, similar to Generous TFT"""

    name = 'Stochastic WSLS'

    def __init__(self, ep=0.05):
        self.ep = ep
        four_vector = (1.-ep, ep, ep, 1.-ep)
        super(self.__class__, self).__init__(four_vector)


class ZeroDeterminantPlayer(MemoryOnePlayer):
    """Abstraction for ZD players. The correct formula is Equation 14 in
    http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0077886 .
    These players enforce a linear difference in stationary payoffs
    s * (S_xy - l) = S_yx - l, yielding extortionate strategies with l = P and
    generous strategies when l = R and s > 0"""
    name = 'ZD ABC'

    def __init__(self, phi=0., s=None, l=None):
        (R, P, S, T) = Game().RPST()
        if s is None:
            s = 1
        if l is None:
            l = R

        # Check parameters
        s_min = - min((T-l) / (l-S), (l-S) / (T-l))
        if (l < P) or (l > R) or (s > 1) or (s < s_min):
            raise ValueError

        p1 = 1 - phi * (1 - s) * (R - l)
        p2 = 1 - phi * (s * (l - S) + (T - l))
        p3 = phi * ((l - S) + s * (T - l))
        p4 = phi * (1 - s) * (l - P)

        four_vector = [p1, p2, p3, p4]
        MemoryOnePlayer.__init__(self, four_vector)


class ZDGTFT2(ZeroDeterminantPlayer):
    """A Generous Zero Determinant Strategy."""

    name = 'ZD-GTFT-2'

    def __init__(self, phi=0., chi=2.):
        (R, P, S, T) = Game().RPST()
        ZeroDeterminantPlayer.__init__(self, phi=0.25, s=0.5, l=R)


class ZDExtort2(ZeroDeterminantPlayer):
    """An Extortionate Zero Determinant Strategy."""

    name = 'ZD-Extort-2'

    def __init__(self):
        (R, P, S, T) = Game().RPST()
        ZeroDeterminantPlayer.__init__(self, phi=1./9, s=0.5, l=P)


### Strategies for recreating Axelrod's tournament ###


class Grofman(MemoryOnePlayer):
    """
    Cooperates with probability 2/7.
    """

    name = "Grofman"

    def __init__(self):
        p = float(2) / 7
        four_vector = (p, p, p, p)
        super(self.__class__, self).__init__(four_vector)


class Joss(MemoryOnePlayer):
    """
    Cooperates with probability 0.9 when the opponent cooperates, otherwise 
    emulates Tit-For-Tat.
    """

    name = "Joss"

    def __init__(self, p=0.9):
        four_vector = (p, 0, p, 0)
        super(self.__class__, self).__init__(four_vector)


class SoftJoss(MemoryOnePlayer):
    """
    Defects with probability 0.9 when the opponent defects, otherwise 
    emulates Tit-For-Tat.
    """

    name = "Soft Joss"

    def __init__(self, q=0.9):
        four_vector = (1., 1 - q, 1, 1 - q)
        super(self.__class__, self).__init__(four_vector)



from axelrod import Player
import random


class Random(Player):
    """A player who randomly chooses between cooperating and defecting."""

    name = 'Random'
    memory_depth = 0  # Memory-one Four-Vector = (p, p, p, p)

    def __init__(self, p=0.5):
        Player.__init__(self)
        self.p = p

    def strategy(self, opponent):
        r = random.random()
        if r > self.p:
            return 'D'
        return 'C'



from axelrod import Player


class BackStabber(Player):
    """
    Forgives the first 3 defections but on the fourth
    will defect forever. Defects on the last 2 rounds unconditionally.
    """

    name = 'BackStabber'
    memory_depth = float('inf')  # Long memory
    stochastic = False

    def strategy(self, opponent):
        if not opponent.history:
            return 'C'
        if len(opponent.history) > (self.tournament_attributes['length'] - 3):
            return 'D'
        if opponent.defections > 3:
            return 'D'
        return 'C'


class DoubleCrosser(Player):
    """
    Forgives the first 3 defections but on the fourth
    will defect forever. If the opponent did not defect
    in the first 6 rounds the player will cooperate until
    the 180th round. Defects on the last 2 rounds unconditionally.
    """

    name = 'DoubleCrosser'
    memory_depth = float('inf')  # Long memory
    stochastic = False

    def strategy(self, opponent):
        cutoff = 6

        if not opponent.history:
            return 'C'
        if len(opponent.history) > (self.tournament_attributes['length'] - 3):
            return 'D'
        if len(opponent.history) < 180:
            if len(opponent.history) > cutoff:
                if 'D' not in opponent.history[:cutoff + 1]:
                    if opponent.history[-2:] != ['D', 'D']:  # Fail safe
                        return 'C'
        if opponent.defections > 3:
            return 'D'
        return 'C'



import copy
import inspect

from axelrod import Player, RoundRobin, Game, update_histories


def simulate_match(player_1, player_2, strategy, rounds = 10):
    """Simulates a number of matches."""
    for match in range(rounds):
        play_1, play_2 = strategy, player_2.strategy(player_1)
        # Update histories and counts
        update_histories(player_1, player_2, play_1, play_2)

def roll_back_history(player, rounds):
    """Undo the last `rounds` rounds as sufficiently as possible."""
    for i in range(rounds):
        play = player.history.pop(-1)
        if play == 'C':
            player.cooperations -= 1
        elif play == 'D':
            player.defections -= 1

def look_ahead(player_1, player_2, rounds=10):
    """Looks ahead for `rounds` and selects the next strategy appropriately."""
    results = []
    game = Game()

    # Simulate plays for `rounds` rounds
    strategies = ['C', 'D']
    for strategy in strategies:
        #opponent_ = copy.deepcopy(player_2) # need deepcopy here
        opponent_ = player_2
        round_robin = RoundRobin(players=[player_1, opponent_], game=game,
                                 turns=rounds)
        simulate_match(player_1, opponent_, strategy, rounds)
        results.append(round_robin._calculate_scores(player_1, opponent_)[0])

        # Restore histories and counts
        roll_back_history(player_1, rounds)
        roll_back_history(player_2, rounds)

    return strategies[results.index(max(results))]


class MindReader(Player):
    """A player that looks ahead at what the opponent will do and decides what to do."""

    name = 'Mind Reader'
    stochastic = True # Don't cache me

    def strategy(self, opponent):
        """Pretends to play the opponent a number of times before each match.
        The primary purpose is to look far enough ahead to see if a defect will
        be punished by the opponent.
        If the MindReader attempts to play itself (or another similar strategy),
        then it will cause a recursion loop, so this is also handeled in this
        method, by defecting if the method is called by strategy
        """

        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        calname = calframe[1][3]

        if calname in ('strategy', 'simulate_match'):
            return 'D'

        best_strategy = look_ahead(self, opponent)

        return best_strategy


class ProtectedMindReader(MindReader):
    """A player that looks ahead at what the opponent will do and decides what to do.
    It is also protected from mind control strategies"""

    name = 'Protected Mind Reader'

    def __setattr__(self, name, val):
        """Stops any other strategy altering the methods of this class """

        if name == 'strategy':
            pass
        else:
            self.__dict__[name] = val



from axelrod import Player

flip_dict = {'C': 'D', 'D': 'C'}


class TitForTat(Player):
    """A player starts by cooperating and then mimics previous move by opponent."""

    name = 'Tit For Tat'
    memory_depth = 1  # Four-Vector = (1.,0.,1.,0.)

    @staticmethod
    def strategy(opponent):
        return 'D' if opponent.history[-1:] == ['D'] else 'C'


class TitFor2Tats(Player):
    """A player starts by cooperating and then defects only after two defects by opponent."""

    name = "Tit For 2 Tats"
    memory_depth = 2  # Long memory, memory-2

    @staticmethod
    def strategy(opponent):
        return 'D' if opponent.history[-2:] == ['D', 'D'] else 'C'


class TwoTitsForTat(Player):
    """A player starts by cooperating and replies to each defect by two defections."""

    name = "Two Tits For Tat"
    memory_depth = 2  # Long memory, memory-2

    @staticmethod
    def strategy(opponent):
        return 'D' if 'D' in opponent.history[-2:] else 'C'


class Bully(Player):
    """A player that behaves opposite to Tit For Tat, including first move.

    Starts by defecting and then does the opposite of opponent's previous move.
    This the complete opposite of TIT FOR TAT, also called BULLY in literature.
    """

    name = "Bully"
    memory_depth = 1  # Four-Vector = (0.,1.,0.,1.)

    @staticmethod
    def strategy(opponent):
        return 'C' if opponent.history[-1:] == ['D'] else 'D'


class SneakyTitForTat(Player):
    """Tries defecting once and repents if punished."""

    name = "Sneaky Tit For Tat"
    memory_depth = float('inf')  # Long memory

    def strategy(self, opponent):
        if len(self.history) < 2:
            return "C"
        if 'D' not in opponent.history:
            return 'D'
        if opponent.history[-1] == 'D' and self.history[-2] == 'D':
            return "C"
        return opponent.history[-1]


class SuspiciousTitForTat(Player):
    """A TFT that initially defects."""

    name = "Suspicious Tit For Tat"
    memory_depth = 1  # Four-Vector = (1.,0.,1.,0.)

    @staticmethod
    def strategy(opponent):
        return 'C' if opponent.history[-1:] == ['C'] else 'D'


class AntiTitForTat(Player):
    """A strategy that plays the opposite of the opponents previous move.
    This is similar to BULLY above, except that the first move is cooperation."""

    name = 'Anti Tit For Tat'
    memory_depth = 1  # Four-Vector = (0.,1.,0.,1.)

    @staticmethod
    def strategy(opponent):
        return 'D' if opponent.history[-1:] == ['C'] else 'C'


class HardTitForTat(Player):
    """A variant of Tit For Tat that uses a longer history for retaliation."""

    name = 'Hard Tit For Tat'
    memory_depth = 3  # memory-three

    @staticmethod
    def strategy(opponent):
        # Cooperate on the first move
        if not opponent.history:
            return 'C'
        # Defects if 'D' in the opponent's last three moves
        if 'D' in opponent.history[-3:]:
            return 'D'
        # Otherwise cooperates
        return 'C'


class HardTitFor2Tats(Player):
    """A variant of Tit For Two Tats that uses a longer history for
    retaliation."""

    name = "Hard Tit For 2 Tats"
    memory_depth = 3  # memory-three

    @staticmethod
    def strategy(opponent):
        # Cooperate on the first move
        if not opponent.history:
            return 'C'
        # Defects if two consecutive 'D' in the opponent's last three moves
        history_string = "".join(opponent.history[-3:])
        if 'DD' in history_string:
            return 'D'
        # Otherwise cooperates
        return 'C'
